# bash range expand library 0.10
# (C) Copyright  Heinz Deinhart 2015-2018
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# This is a bash library, it has to be sourced not called.
#
# Written for bash < 4.4 and thus not using associative array references.
# In future version this will be added.
# Currently that means that macros will have to reside in a well known name
# (BRE_RANGE_MACROS).

[ "$0" != "$BASH_SOURCE" ] || {
	echo "$0: this script must be sourced, not called" >&2
	exit 1
}

declare -A BRE_RANGE_MACROS

#@@ range expand

#@ DEPRECATED: bre_expand_ranges <prefix> <suffix> <reserved> <input-string> [..]
# This function is deprecated, use bre_expand instead.
#  <prefix> and <suffix> are strings without spaces
#  <reserved> is reserved for the name of the macro associative array
#             this will be implemented in a future version
#  <input-string> can be the following:
#    a macro: a string which is key in the associative array BRE_RANGE_MACROS
#             its value will be expanded in recursive fashion
#    a range (<number>-<number>): will be expanded to each member of range
#                                 prefix and suffix will be added to each
#    a number: prefix and suffix will be added
#    any other string (must not contain spaces): will be copied unmodified
# errors: 2
function bre_expand_ranges() {
    bb_msg warning "bre_expand_ranges is deprecated, switch to new API"

    (( $# > 2 )) || {
	bb_ierr "bre_expand_ranges: called with $# instead of 3 or more args"
	return 2
    }

    local _prefix="$1"
    local _suffix="$2"
    # local reserved="$3"
    shift 3

    local _result=""
    while (( $# > 0 ))
    do
	_result="$_result $(_bre_parse_range_old "$_prefix" "$_suffix" "" "$1")"
        shift
    done

    echo "${_result# }"
}

#@ bre_expand <ret-array-vn> <range-dict-vn> <prefix> <suffix> <range> [..]
# expand ranges found in range-dict and return it in ret-array
# range-dict keys are names of ranges (whitespace not allowed)
#   and their values can be:
#     a macro: any text string that matches a key of range-dict
#              its value will be expanded in recursive fashion
#     a range (<number>-<number>): will be expanded to each member of range
#                                  prefix and suffix will be added to each
#     a number: prefix and suffix will be added
#     any other string will be copied unmodified
# errors: 2, 3
function bre_expand() {
    (( $# > 3 )) || {
	bb_ierr "bre_expand: called with $# instead of 4 or more args"
	return 2
    }

    local _ret_vn="$1"
    local _range_dict_vn="$2"
    local _prefix="$3"
    local _suffix="$4"
    shift 4

    bb_validate_vn "bre_expand:ret-array-vn" "$_ret_vn" || return 3
    bb_validate_vn "bre_expand:range-dict-vn" "$_range_dict_vn" || return 3

    local -a bre_expand__range_dict_keys

    # shellcheck disable=SC2034
    local -a bre_expand__range_dict_keys
    bb_dict_keys bre_expand__range_dict_keys "$_range_dict_vn"

    eval "$_ret_vn=()"

    while (( $# > 0 ))
    do
	_bre_parse_range "$_ret_vn" "$_range_dict_vn" \
	    bre_expand__range_dict_keys "$_prefix" "$_suffix" "$1" || \
	    bb_ierr "bre_expand: failed to call _bre_parse_range: $?"
	shift
    done
}

### internal functions below ###

# _bre_parse_range <add-to-array-vn> <range-dict-vn> <range-dict-keys-vn>
#                  <prefix> <suffix> <range>
function _bre_parse_range() {
    (( $# == 6 )) || {
	bb_ierr "_bre_parse_range called with $# instead of 6 args"
	return 2
    }

    local _add_to_array_vn="$1"
    local _range_dict_vn="$2"
    local _range_dict_keys_vn="$3"
    local _prefix="$4"
    local _suffix="$5"
    local _range="$6"

    # check if range is a macro
    if bb_array_contains_value "$_range_dict_keys_vn" "$_range"
    then
	local _ref _macro _name

	_ref="${_range_dict_vn}[$_range]"
	_macro="${!_ref}"

	for _name in $_macro # whitespace separated
	do
	    _bre_parse_range "$_add_to_array_vn" "$_range_dict_vn" \
		"$_range_dict_keys_vn" "$_prefix" "$_suffix" "$_name" || \
		bb_ierr "_bre_parse_range: failed to call _bre_parse_range: $?"
	done

	return 0
    fi

    # range is no macro
    local -a _expansion
    local -i _i
    # expand range
    case "$_range" in
	"") return 0;; # ignore empty entries
        *-*-*) _expansion=("$_range");; # too many dashs, that is not a range
        *[!0-9-]*) _expansion=("$_range");; # that is not a range, too
        *-*) # range of numbers
	    _expansion=()
	    for (( _i="${_range%-*}"; _i <= "${_range#*-}"; _i++ ))
	    do
                _expansion+=("$_prefix$_i$_suffix")
            done
            ;;
        *) _expansion=("$_prefix$_range$_suffix") # just a number
    esac

    # add result to add-to-array
    eval "$_add_to_array_vn+=(\"\${_expansion[@]}\")"

    return 0
}

# _bre_parse_range <prefix> <suffix> <reserved> <range>
function _bre_parse_range_old() {
    (( $# == 4 )) || {
	bb_ierr "_bre_parse_range_old called with $# instead of 4 args"
	return 2
    }

    local prefix="$1"
    local suffix="$2"
    # local reserved="$3"
    local range="$4"

    # check if macros are defined
    if [[ "${BRE_RANGE_MACROS[@]-unset}" != unset ]]
    then
	# check if range is a macro
	if [ "${BRE_RANGE_MACROS[$range]+isset}" == isset ]
	then
	    bre_expand_ranges "$prefix" "$suffix" "" \
		${BRE_RANGE_MACROS[$range]}
            return 0
	fi
    fi

    # expand range
    case "$range" in
        *-*-*) # that is not a range
            echo "$range"
            ;;
        *[!0-9-]*) # that is not a range, too
            echo "$range"
            ;;
        *-*) # real range
	    local result=""
            for n in $(seq "${range%-*}" "${range#*-}")
            do
                result="$result ${prefix-}$n${suffix-}"
            done
	    echo "${result# }"
            ;;
        *) # just a number
            echo "${prefix-}$range${suffix-}"
    esac

    return 0
}
